import logging
import requests
import json
import pandas as pd
import os
from datetime import datetime
from SPARQLWrapper import SPARQLWrapper, JSON

from .models import DataAccessObject

token_var_name = 'SEADATANET_BEACON_API_TOKEN'


def to_seadatanet_datetime(dt:datetime):
    '''converts a Python datetime to the Seadatanet format, i.e. 2007-09-01'''
    return dt.strftime('%Y-%m-%d')

def _format_seadatanet_query_body(mindate, maxdate, minlon, maxlon, minlat, maxlat, mindepth, maxdepth, matching_p01s: list[str]):
    query_parameters = []
    query_parameters.append("LONGITUDE")
    query_parameters.append("LATITUDE")
    query_parameters.append("TIME")
    query_parameters.append({"function": "coalesce", "args": ["DEPTH", "PRES"], "alias": "DEPTH"})
    query_parameters.append("SDN_STATION")
    query_parameters.append("SDN_EDMO_CODE")
    query_parameters.append("SDN_LOCAL_CDI_ID")
    for p01 in matching_p01s:
        query_parameters.append(p01)

    filters = []
    filters.append({"for_query_parameter": "TIME", "min": f"{mindate}T00:00:00", "max": f"{maxdate}T00:00:00"})
    filters.append({"for_query_parameter": "DEPTH", "min": mindepth, "max": maxdepth})
    filters.append({"for_query_parameter": "LONGITUDE", "min": minlon, "max": maxlon})
    filters.append({"for_query_parameter": "LATITUDE", "min": minlat, "max": maxlat})


    parameter_or_filters = []
    for p01 in matching_p01s:
        parameter_or_filters.append({"is_not_null": {"for_query_parameter": p01}})
    or_filter = {
        "or" : parameter_or_filters
    }
    filters.append(or_filter)
    body = {
        "query_parameters": query_parameters,
        "filters": filters,
        "output": {"format": "ipc"},
    }
    return body

class SeadatanetError(Exception):
    def __init__(self, message):            
        super().__init__(message)

class AuthenticationTokenError(SeadatanetError):
    def __init__(self, message):            
        super().__init__(message)
        logging.info('To request a token, go down at Paul\'s at maris.nl and tell him you want the Bob Marley extra crispy  -> maris.nl/contact')
        
class SeaDataNetBeaconDAO(DataAccessObject):
    
    def __init__(self, token = None, beacon_url = "https://beacon-cdi.maris.nl/api/query", vocabulary_endpoint_url = 'https://vocab.nerc.ac.uk/sparql/sparql'):
        self.token = token
        self.__check_token()
        self.vocabulary_endpoint_url = vocabulary_endpoint_url
        self.beacon_url = beacon_url


    def __check_token(self):
        if self.token is None:
            try:
                self.token = os.environ[token_var_name]
            except:
                raise AuthenticationTokenError(f'No Seadatanet authentication token. pass it as a constructor argument or set the env variable {token_var_name}')

    def _get_cdi_input(self):
        responseinfo = requests.get(self.beacon_url + "/available-columns",headers={"Authorization": f"Bearer {self.token}"},)
        return responseinfo.json()
    

    def _get_EXV_iadopt(self, exv_code):
        # SPARQL endpoint
        endpoint_url = self.vocabulary_endpoint_url
        # Construct full identifier
        exv_identifier = f"SDN:EXV::{exv_code}"

        # Create the query with the user input
        sparql_query = f"""
        PREFIX dce: <http://purl.org/dc/elements/1.1/>
        PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
        PREFIX iadopt: <https://w3id.org/iadopt/ont#> 
        PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>

        SELECT DISTINCT ?p01 ?prefLabel ?notation
        WHERE {{
        ?exv a skos:Concept .
        ?exv dce:identifier "{exv_identifier}" .

        OPTIONAL {{?exv iadopt:hasApplicableMatrix ?matrix .}}
        ?exv iadopt:hasApplicableObjectOfInterest ?ooi .
        ?exv iadopt:hasApplicableProperty ?property .

        <http://vocab.nerc.ac.uk/collection/P01/current/> skos:member ?p01 .

        OPTIONAL {{ ?p01 iadopt:hasMatrix ?matrix . }}
        ?p01 iadopt:hasObjectOfInterest ?ooi .
        ?p01 iadopt:hasProperty ?property .

        OPTIONAL {{ ?p01 skos:prefLabel ?prefLabel . }}
        OPTIONAL {{ ?p01 skos:notation ?notation . }}
        }}
        """
        # Set up the SPARQL request
        sparql = SPARQLWrapper(endpoint_url)
        sparql.setQuery(sparql_query)
        sparql.setReturnFormat(JSON)

        # Run the query and parse results
        results = sparql.query().convert()

        codes = []

        # Show results
        for result in results["results"]["bindings"]:
            uri = result.get("p01", {}).get("value", "")
            codes.append(uri.rstrip("/").split("/")[-1])

        params = self._get_cdi_input()
        matching_p01s = []
        for code in codes:
            if code in params:
                matching_p01s.append(code)

        return matching_p01s
    
    def access(self, exv_code, mindate, maxdate, minlon, maxlon, minlat, maxlat, mindepth, maxdepth)->pd.DataFrame:
        '''
        Access method for seadatanet data exposed via the BEACON API
        
        Parameters
        ----------
        exv_code:str
            the Essential Climate Variable you want to retrieved, expressed with 
            a alphanumeric code.
            Example: "EXV017"
        
        mindate:datetime
            time interval lower bound
        
        maxdate:datetime 
            time interval upper bound
        
        minlon, 
        
        maxlon, 
        
        minlat, 
        
        maxlat, 
        
        mindepth, 
        
        maxdepth
        
        '''
        matching_p01s = self._get_EXV_iadopt(exv_code)
        query_body = _format_seadatanet_query_body(to_seadatanet_datetime(mindate), to_seadatanet_datetime(maxdate), minlon, maxlon, minlat, maxlat, mindepth, maxdepth, matching_p01s)
        response = requests.post(self.beacon_url, json.dumps(query_body),
                    headers={"Authorization": f"Bearer {self.token}",
                                "Content-type": "application/json"})
        response.raise_for_status() # this raises errors in case of 40X or 50X
        if response.status_code == 204:
            logging.warning("No data has been found for your query")
            return None
        # otherwise, let's convert it to PANDAS
        df = pd.read_feather(response.content)
        df = df.set_index("TIME").sort_index() 
        cols_to_exclude = ['SDN_STATION', 'SDN_EDMO_CODE', 'SDN_LOCAL_CDI_ID']
        df[df.columns.difference(cols_to_exclude)] = df[df.columns.difference(cols_to_exclude)].apply(pd.to_numeric, errors='coerce')
        df[exv_code] = df[matching_p01s].mean(axis=1)
        df.drop(columns=matching_p01s, inplace=True)
        return df

    



from abc import ABC, abstractmethod
import logging
import os
import requests
from prance import ResolvingParser


from .utils import call_restful_api
from ..decorators import delegates

class DataAccessObject(ABC):
    '''abstract class for any object that allows to access data in the ENVRI HUB'''

    @abstractmethod
    def access(self, **kwargs):
        pass

class FileAccessObject(DataAccessObject):

    def __init__(self, url):
        self.url = url

    @delegates(requests.get)
    def access(self, stream = True, **kwargs)->bytes:
        '''Allows to retrieve the file contents as bytes'''
        return requests.get(self.url, stream=stream, **kwargs).content
    
    def download(self, path='.tempfile', stream = True, **kwargs):
        '''
        Dowloads the file to a specified file system location

        Paramters
        ---------
        path:str defualt:".tempfile"
            a file system location to store the file

        '''
        response = requests.get(self.url, stream=stream, **kwargs)
        try:
            # the "good" path: stuff is wrapped in a http response with shit and giggles
            if "content-disposition" in response.headers:
                content_disposition = response.headers["content-disposition"]
                if content_disposition in ['inline', 'attachment']:
                    filename = path
                else:
                    filename = content_disposition.split("filename=")[1]
            else:
                filename = path
            with open(filename, mode="wb") as file:
                file.write(response.content)
                logging.info(f"Downloaded file {filename}")
        except:
            # the "wtf" path: the response is just a string
            with open(path, mode="wb") as file:
                file.write(response)
                logging.info(f"Downloaded file {filename}")
            



class RESTfulAPIDataAccess(DataAccessObject):

    def __init__(self, base_address):
        self.base_address = base_address
        
    @delegates(call_restful_api)
    def access(self, **kwargs):
        return call_restful_api(self.base_address, **kwargs)

class RESTfulAPIwithAPIKeyDataAccess(RESTfulAPIDataAccess):

    def __init__(self, base_address, key_name):
        self._service_key = os.environ.get(key_name)
        if self._service_key is None:
            logging.warning(f'API key {key_name} not found in global environment')
        

class OpenAPI3DataAccess(RESTfulAPIDataAccess):
    
    def __init__(self, **kwargs):
        super().__init__(self.servers.preferred_server)
    
    def access(self, method, **kwargs):
        return getattr(self, method)(**kwargs)
        


## Datetime parsing
from datetime import datetime
import pytz
from envrihub.data_access.argo import to_argo_datetime
from envrihub.data_access.open_api_client import _to_parameter_name
from envrihub.data_access.seadatanet import to_seadatanet_datetime


def test_datetime_parsing_argo_outformat():
    t = to_argo_datetime(datetime(2024, 1, 23, 9, 21, 54, tzinfo=pytz.utc))
    target = '2024-01-23T09:21:54+00:00'
    assert t==target, f'Incorrectly parsed datetime, expected {target}, but got {t}'

def test_datetime_parsing_seadatanet_outformat():
    t = to_seadatanet_datetime(datetime(2024, 1, 23, 9, 21, 54, tzinfo=pytz.utc))
    target = '2024-01-23'
    assert t==target, f'Incorrectly parsed datetime, expected {target}, but got {t}'

def test_to_parameter_name():
    assert _to_parameter_name('Content-Type') == 'Content_Type', 'Parameter names are not not sanitized enough'


import inspect
import io
import os
import logging
import posixpath
from urllib.parse import urljoin

import requests

from .models import OpenAPI3DataAccess
from .utils import call_restful_api

from prance import ResolvingParser

def recursion_limit_handler_void(limit, refstring, recursions):
    return {}

# the noice factory method
def build_web_service_client(service_documentation, resource_name=None)->type:
    parser = ResolvingParser(service_documentation, recursion_limit_handler=recursion_limit_handler_void)
    parser.parse()
    specification = parser.specification
    return  build_access_class(specification, resource_name)

class ApplicationServer():
    def __init__(self, url, description):
        self.url = url
        self.description = description
    
    def call_function(self, function_name, **kwargs):
        service_url =self.url + function_name
        return call_restful_api(url=service_url, **kwargs)
    

class ApplicationServerPool():
    def __init__(self, specification_servers = [], base_address = ''):
        self.servers = []
        self.preferred_server = None
        self._dead_servers = set()
        # import appservers from the OpenAPI specification
        if specification_servers is not None:
            # beware the base address here: some openapi specs have only
            # relative addresses, so with urljoin we try to combine them
            # with an absolute address. Urljoin is smart enough to tell
            # if the server url is already an absolute one and in such a 
            # case it ignores the base address.
            self.servers = [ApplicationServer(urljoin(base_address, server.get('url')), server.get('description')) for server in specification_servers]
            self.preferred_server = self.servers[0] if len(self.servers) > 0 else None
            self.check_servers()
        
    def call_function(self, **kwargs):
        unavailable_servers = set()
        try:
            return self.preferred_server.call_function(**kwargs)
        except requests.exceptions.HTTPError as e:
            unavailable_servers.add(self.preferred_server)
            # find an available server
            response = None
            for s in self.servers:
                if (s not in self._dead_servers) and (s not in unavailable_servers) :
                    try:
                        response = s.call_function(**kwargs)
                        self.preferred_server = s
                        return response
                    except:
                        unavailable_servers.add(s)
            logging.error('Request to ' + kwargs.get('function_name') + ' cannot be satisfied on any server')    
                        
    def check_servers(self):
        for s in self.servers:
            resp = requests.get(s.url)
            if  200 <= resp.status_code < 300:
                logging.info(f'{s.url} is online')
            else:
                self._dead_servers.add(s)
                logging.warning(f'{s.url} fails with HTTP status {resp.status_code}')
            
def _to_parameter_name(par_name):
    '''Given a string, it returns a version of it that can be used as a Python parameter name'''
    return par_name.translate({ord(c): '_' for c in '/!@#*+-%&|[]()$ \r\n\t'})

def _remove_duplicates(elements:list[inspect.Parameter]):
    '''monkey-patch solution to remove duplicate entries from the
    signature parameters list without destroying its ordering'''
    keys = set()
    out_els = []
    for e in elements:
        if e.name not in keys:
            out_els.append(e)
            keys.add(e.name)
    return out_els

def _build_translation_tables(*args):
    '''given any collection of strings, it evaluates sanitized versions of said strings
     and returns two dictionaries containing the translation mapping:
     original string -> sanitized string
     and
     sanitized string -> original string'''
    ab = {}
    ba = {}
    for d in args:
        for k in d:
            san_k = _to_parameter_name(k)
            ab[k] = san_k
            ba[san_k] = k
    return ab, ba

def build_caller_function(function_name, method, params, route_parameters, content_parameters, default_headers, request_content_type, response_content_type,
                         docstring=''):
    '''Creates a class method to call a REST service.
    The otuput callable object must be placed inside an object with
    a *servers* attribute that exposes a *call_function* method, because the generated callable
    will only parse parameters and send them to that function in that attribute'''
    # let's build a translation dictionary for parameter names
    # we need this because stuff like 'Content-Type' is legit HTTP parameter name
    # but it wuld not work as a Python variable name, hence the need for
    # a name conversion
    par_ab, par_ba = _build_translation_tables(params, route_parameters, content_parameters, default_headers)
    # in the method signature, we'll put the *sanitized* version of the names (so par_ab[par_name])
    # but then in the actual request dictionary, we should send the original parameter name (so par_ba[par_name])
    # gogadget closure
    def caller_function(self, **kwargs): # the only way to dynamically build a method's signature is to use **kwargs and then do some deep black magic with inspect
        # build request parameters
        # query/form parmaters
        request_paramaters = {p:(kwargs[par_ab.get(p)] if (par_ab.get(p) in kwargs) else v) for p, v in params.items()}
        # route parameters
        path_params = {p:(kwargs[par_ab.get(p)] if (par_ab.get(p) in kwargs) else v) for p, v in route_parameters.items()}
        # the idea here is that if the user wants to send a payload they manually built, they always can
        if 'data' in kwargs:
            request_payload = kwargs.get('data')
        else: # otherwise we'll cobble one together from the specification
            request_payload = {p:(kwargs[par_ab.get(p)] if (par_ab.get(p) in kwargs) else v) for p, v in content_parameters.items()}
        # build request headers
        #print(request_payload)
        request_headers = {p:(kwargs[par_ab.get(p)] if (par_ab.get(p) in kwargs) else v) for p, v in default_headers.items()}
        # now the ApplicationServerPool object should handle the request
        return self.servers.call_function(function_name = function_name, data=request_payload, params = request_paramaters,
                                          path_params=path_params, headers = request_headers,
                                files = kwargs.get('files'), method=method, additional_headers= kwargs.get('additional_headers'),
                                payload_content_type =request_content_type,
                                expected_response_content_type = response_content_type)
    # adding docstring
    caller_function.__doc__ = docstring
    # and signature
    signature_params = [inspect.Parameter(name="self", kind = inspect.Parameter.POSITIONAL_ONLY)]
    signature_params = signature_params + [inspect.Parameter(name=par_ab.get(p), default = v, kind = inspect.Parameter.KEYWORD_ONLY) for p, v in params.items()]
    signature_params = signature_params + [inspect.Parameter(name=par_ab.get(p), default = v, kind = inspect.Parameter.KEYWORD_ONLY) for p, v in route_parameters.items()]
    signature_params = signature_params + [inspect.Parameter(name=par_ab.get(p), default = v, kind = inspect.Parameter.KEYWORD_ONLY) for p, v in content_parameters.items()]
    signature_params = signature_params + [inspect.Parameter(name=par_ab.get(p), default = v, kind = inspect.Parameter.KEYWORD_ONLY) for p, v in default_headers.items()]
    signature_params.append(inspect.Parameter(name='data', default = None, kind = inspect.Parameter.KEYWORD_ONLY))
    signature_params.append(inspect.Parameter(name='files', default = None, kind = inspect.Parameter.KEYWORD_ONLY))
    signature_params.append(inspect.Parameter(name='additional_headers', default = None, kind = inspect.Parameter.KEYWORD_ONLY, annotation=dict))
    # soooo, sometimes some paramters may appear twice becasue of weird specifciations...
    # therefore we need to turn signature params into a set and then back into a list 
    # signature_params= list(set(signature_params))
    # UPDATE: that was a very bad idea, becasue it messed up parameters order
    # so in the end I had to write a fugly monkey patch method to fix the problem
    signature_params = _remove_duplicates(signature_params)
    #print(signature_params)
    caller_function.__signature__ = inspect.Signature(parameters=signature_params)
    return caller_function

def _sanitize(url_string):
    '''Removes characters that cannot be into a function's name'''
    return url_string.translate({ord(c): None for c in '/!@#*+-%&|[]{}()$ \r\n\t'})

def parse_parameters(parameter_list:list):
    '''parses the contents object in the OpenAPI specification
    to extrace content parameters and their default values. Plus docstrings'''
    params = {}
    path_params = {} # inside the parameters object there may be path parameters too!
    header_params = {} # and stuff that goes in the header too!
    docstring = ''
    for param_dict in parameter_list:
        name = param_dict.get('name')
        param_type = param_dict.get('in') # this is the rat bastard telling where to shove the parameter
        # these are the schema informations
        if isinstance(param_dict.get('schema'), dict):
            description = param_dict.get('schema').get('description')
            default = param_dict.get('schema').get('default')
            par_type = param_dict.get('schema').get('type')
        else:
            description = ''
            default = None
            par_type = 'any'
        # put things in the correct bin
        if param_type == 'header':
            header_params[name] = default
        elif param_type == 'path':
            path_params[name] = default
        else:
            params[name] = default
            # assembling the docstring
        docstring = docstring +  str(name) + ':' + str(par_type) + ' '+ 'default = ' + str(default) + '\n\t' + str(description) + '\n'
    return params, header_params, path_params, docstring
        

def build_access_class(specification:dict, resource_name = None, base_address = '')->type:
    if resource_name is None:
        resource_name = _sanitize(specification.get('info').get('title'))
    class_methods={}
    servers = ApplicationServerPool(specification.get('servers'), base_address)
    # now we have to go through the paths dictionary
    # which is made of abominations like this
    # {
    #  '/address': {'post': {...}, 'get':{...},
    #   ...
    # }
    for url, methods in specification.get('paths').items():
        supported_methods = ','.join([x for x in methods])
        logging.info(f"Operation: {url}, methods: {supported_methods}")
        for method, specs in methods.items():
            # here are the GET/POST/PUT/whatever variants
            # the possible responses are in the 'responses' object, which is an interable
            # likely there will be some error codes and somewhere a 200-OK one
            response_content_type = None
            if specs.get('responses'):
                for response, content in specs.get('responses').items():
                    if response in ['200', '201', '202']:
                        # the content we are looking for! But beware: multiple response formats might be acceptable!
                        response_content_type = []
                        if isinstance(content.get('content'), dict):
                            for content_type, notes in content.get('content').items():
                                response_content_type.append(content_type)
                        # if the list has only one element, then we'll just pass down that one element
                        if len(response_content_type) == 1:
                            response_content_type = response_content_type[0]
                    
            description = specs['description'] if specs.get('description') else ''
            docstring = description + '\n\nParameters\n----------\n'
            # here let's parse the GET http parameters
            params = {}
            path_params = {} # inside the parameters object there may be path parameters too!
            header_params = {} # and stuff that goes in the header too!
            if isinstance(specs.get('parameters'), list):
                params, header_params, path_params, param_docstring = parse_parameters(specs.get('parameters'))
                docstring = docstring + param_docstring
            # now for the headers
            default_headers = {}
            # TODO read specification for headers, they should be in the *security* 
            # section of the open API spec.
            # here, if any, the parameters to be structured into a JSON payload
            default_headers = default_headers | header_params
            content_parameters={}
            request_content_type = None
            if isinstance(specs.get('requestBody'), dict):
                contents = specs.get('requestBody').get('content')
                if isinstance(contents, dict):
                    # let's get through all the stuff you can shove in the request body
                    for request_content_type, content_info in contents.items():
                        default_headers['Content-Type'] = request_content_type
                        content_parameters={}
                        # of course we do that only if we are dealing with a dictionary-like payload
                        if (request_content_type == 'application/json') and (isinstance(content_info.get('schema'), dict)):
                            if isinstance(content_info.get('schema').get('properties'), dict):
                                for name, req_param in content_info.get('schema').get('properties').items():
                                    
                                    description = req_param.get('description')
                                    default = req_param.get('default')
                                    par_type = req_param.get('type')
                                    docstring = docstring +  str(name) + ':' + str(par_type) +' ' + 'default = ' + str(default) + '\n\t' + str(description) + '\n'
                                    content_parameters[name] = default
                    content_slug = request_content_type.split('/')[-1]
                    # format the name method so that it looks like get_function_json
                    method_name = f'{method}_{_sanitize(url)}' if len(methods)>1 else _sanitize(url)
                    method_name = f'{method_name}_{content_slug}' if len(contents)>1 else method_name
                    class_methods[method_name]= build_caller_function(url, method, params, route_parameters=path_params, 
                                                    content_parameters=content_parameters, default_headers=default_headers,
                                                    request_content_type= request_content_type,response_content_type= response_content_type,
                                                    docstring=docstring)    
            else:
                # let's build a method with no content parameters then
                method_name = f'{method}_{_sanitize(url)}' if len(methods)>1 else _sanitize(url)
                class_methods[method_name]= build_caller_function(url, method, params, path_params,content_parameters, default_headers, request_content_type, response_content_type,
                         docstring)
    # now we have the methods in a dictionary
    # we can build the access object as a new dedicated class
    return type(resource_name, (OpenAPI3DataAccess,), {'servers':servers, **class_methods})


    


from abc import ABC, abstractmethod
import inspect
import logging

import requests

from envrihub.data_access.catalogue_metadata_client import build_epos_catalogue_metadata_access
from envrihub.data_access.models import DataAccessObject, FileAccessObject, RESTfulAPIDataAccess
from envrihub.data_access.open_api_client import build_web_service_client
from envrihub.decorators import delegates

# Exception definition bonanza
class DistributionError(Exception):
    def __init__(self, message):            
        super().__init__(message)

class NullDistributionTypeError(DistributionError):
    def __init__(self, message, resource_id):            
        # Call the base class constructor with the parameters it needs
        super().__init__(message)
        logging.error(f'Resource {resource_id} has no type attribute')

class InvalidDistributionTypeError(DistributionError):
    def __init__(self, message, resource_id, type_value):            
        # Call the base class constructor with the parameters it needs
        super().__init__(message)
        logging.error(f'Resource {resource_id} has invalid type attribute {type_value}')

class InvalidServiceDocumentationError(DistributionError):
    def __init__(self, message, resource_id, url):
        super().__init__(message)
        logging.error(f'Resource {resource_id} documentation at {url} cannot be used to create a data access client')


class CoSObject(ABC):
    pass

class Distribution(CoSObject):
    
    def __init__(self, href:str, title:str, id:str, uid:str, description:str = None, serviceDocumentation:str = None, type=None, build_dao=True):
        self.href = href
        self.id = id
        try:
            self.metadata = self._get_details()
            if self.metadata.get('serviceDocumentation') is not None:
                self.service_documentation = self.metadata.get('serviceDocumentation')
            else:
                self.service_documentation = serviceDocumentation
            if self.metadata.get('type') is not None:
                self.type = self.metadata.get('type')
            else:
                self.type = type
        except:
            logging.error(f'Cannot access metadata at {href}')
            self.service_documentation = serviceDocumentation
            self.type = type

        self.title = title
        self.uid = uid
        self.description =description
        # we'll instance this as soon as it would be required
        if build_dao:
            self._build_dao()
        else:
            self.dao = None

    def _get_details(self):
        response_detail = requests.get(self.href)
        return response_detail.json()
    
    def _build_dao(self):
        if self.type is None:
            raise NullDistributionTypeError('Null type attribute in current Distribution', self.uid)
        elif self.type == 'WEB_SERVICE':
            service_provider = self.metadata.get('serviceProvider')
            # if it is a well known service provider with custom built classes, we'll use them
            # TODO: the case/switch
            declared_parameters = self.metadata.get('serviceParameters')
            base_url = self.metadata.get('endpoint')
            # if people bothered to describe parameters, we'll go with them
            if (base_url is not None) and (declared_parameters is not None):
                # fetch the application type
                classname = self.metadata.get('serviceName').translate({ord(c): None for c in '/!@#*+-%&|[]()$ \r\n\t'})
                content_type=None
                afs = self.metadata.get('availableFormats')
                for af in afs:
                    content_type = af.get('format')
                try:
                    self.dao = build_epos_catalogue_metadata_access(base_url, declared_parameters, content_type,self.description, classname)(base_url)
                except ValueError as ve:
                    logging.warning(ve)
                    logging.warning('Parameters names do not respect Python naming conventions, instantiating a generic REST client')
                    self.dao = RESTfulAPIDataAccess(base_url)
            # otherwise, it's OpenAPI
            else:
                try:
                    self.dao = build_web_service_client(self.service_documentation)()
                except Exception as e:
                    raise InvalidServiceDocumentationError(e, self.uid, self.service_documentation)
        elif self.type == 'DOWNLOADABLE_FILE':
            self.dao = FileAccessObject(self.metadata.get('downloadURL'))
        else:
            raise InvalidDistributionTypeError('Null type attribute in current Distribution', self.uid, self.type)

    @classmethod
    def from_dict(cls, env):
        '''This is mostly used to build the object from the output returned by Web Services'''
        return cls(**{
                k: v for k, v in env.items() 
                if k in inspect.signature(cls).parameters
            })
    
    def access(self, **kwargs):
        return self.dao.access(**kwargs)
    
    def get_dato(self):
        return self.dao
    
    def is_downloadable(self):
        return self.type == 'DOWNLOADABLE_FILE'
    


